const Logger = require('../utils/Logger');
const Utils = require('../utils/Utils');
const { ScraperError } = require('../../errors');
const Show = require('../media/Show');
const Movie = require('../media/Movie');
const Media = require('../media/Media');

class KissAsianScraper {
  constructor(page) {
    this.page = page;
  }

  async load() {
    await this.page.goto(KissAsianScraper.BASE_URL);
    await this.page.waitForSelector(KissAsianScraper.DEFAULTS.SITE_SELECTOR);

    Logger.info('Cloudflare DDOS Protection bypassed, redirected to main site.');
  }

  async finish(browser) {
    await this.page.close();
    await browser.close();
    Logger.info('Closed browser instance.');
  }

  async authenticate(credentials) {
    await this.page.click('#topHolderBox a[href="/Login"]');
    await this.page.waitForSelector('#btnSubmit');

    Logger.info('Authenticating...');

    await this.page.$eval('#username', (el, credentials) => el.value = credentials.username, credentials);
    await this.page.$eval('#password', (el, credentials) => el.value = credentials.password, credentials);

    await this.page.click('#btnSubmit');
    await this.page.waitForSelector('#head a.logo');

    Logger.success(`Successfully logged in as ${credentials.username}.`);
  }

  async getMediaData(mediaURL) {
    await this.page.goto(mediaURL);
    await this.page.waitForSelector(KissAsianScraper.DEFAULTS.SITE_SELECTOR);

    Logger.info(`Visited ${mediaURL}`);

    const name = await this.getCurrentMediaName();
    const type = await this.getCurrentMediaType();
    const isCompleted = await this.getCurrentMediaStatusCompleted();

    if (!isCompleted) {
      Logger.warn(`Warning! ${name} is still not completed. Proceed only if you're certain.`);
    }

    Logger.info(`Getting media links for ${type} ${name}...`);

    const mediaURLs = await this.page.evaluate(() => {
      const table = document.querySelector('.episodeList table.listing');
      const tableBody = table.firstElementChild;
  
      const links = [];
      for (let i = tableBody.children.length - 1; i > 1; i--) {
        const node = tableBody.children.item(i);
        const aElement = node.firstElementChild.firstElementChild;
        links.push(aElement.href);
      }
  
      return links;
    });

    Logger.success(`Scraped ${mediaURLs.length} media links.`);

    if (type === KissAsianScraper.MEDIA_TYPES.SHOW) {
      return new Show(name, 1, mediaURLs);
    }

    const year = await this.getCurrentMediaYear();

    return new Movie(name, year, mediaURLs);
  }

  async populateMediaDownloadURLs(media) {
    if (!(media instanceof Media)) {
      throw new TypeError('Invalid Media instance passed to scraper!');
    }

    const downloadURLs = await Utils.mapSeries(media.mediaURLs, async(mediaURL) => {
      await this.directlyVisitPlayerURL(mediaURL);
      return this.getDownloadLinkForCurrentPlayer();
    });

    media.setDownloadURLs(downloadURLs);
  }

  async directlyVisitPlayerURL(playerURL) {
    await this.page.goto(`${playerURL}&s=mp`);
    await this.page.waitForSelector(KissAsianScraper.DEFAULTS.SITE_SELECTOR);

    const isCaptchaEnabled = await this.page.evaluate(() => {
      const form = document.getElementById('formVerify');

      return !!form;
    });

    if (isCaptchaEnabled) {
      throw new ScraperError(`MP server not available for ${playerURL}`);
    }

    Logger.info(`Directly visited ${playerURL}`);
  }

  async getDownloadLinkForCurrentPlayer() {
    return await this.page.evaluate(() => {
      const divDownload = document.getElementById('divDownload');
      const aElement = divDownload.lastElementChild;
      return aElement.href;
    });
  }

  async skipCaptcha(playerURL) {
    Logger.warn('Skipping captcha...');

    const [skipCaptchaButton] = await this.page.$x('//a[contains(., "FE server")]');

    if (!skipCaptchaButton) {
      throw new ScraperError(`Skip captcha button not found for ${playerURL}`);
    }

    await skipCaptchaButton.click();
    await this.page.waitForSelector(KissAsianScraper.DEFAULTS.SITE_SELECTOR);

    Logger.success('Successfully skipped captcha.');
  }

  async selectEpisodeServerForCurrentPlayer(server) {
    try {
      await this.page.select('#selectServer', server);
      await this.page.waitForSelector(KissAsianScraper.DEFAULTS.SITE_SELECTOR);

      Logger.info(`Changed streaming server to ${server}.`);
    } catch (err) {
      throw new ScraperError(`Server ${server} is unavailable.`);
    }
  }

  async getCurrentMediaName() {
    return await this.page.evaluate(() => {
      const a = document.querySelector('div.barContent a.bigChar');
      return a.text;
    });
  }

  async getCurrentMediaYear() {
    return await this.page.evaluate(() => {
      const infoSpans = document.querySelectorAll('span.info');
      for (let i = 0; i < infoSpans.length; i++) {
        const span = infoSpans.item(i);

        if (span.textContent === 'Date aired:') {
          const date = span.nextSibling.textContent;
          const [year] = date.match(/\d{4}/);

          return parseInt(year, 10);
        }
      }

      return null;
    });
  }

  async getCurrentMediaStatusCompleted() {
    return await this.page.evaluate(() => {
      const infoSpans = document.querySelectorAll('span.info');
      for (let i = 0; i < infoSpans.length; i++) {
        const span = infoSpans.item(i);

        if (span.textContent === 'Status:') {
          const status = span.nextSibling.textContent;
          
          return status.contains('Completed');
        }
      }

      return null;
    });
  }

  async getCurrentMediaType() {
    const isMovie = await this.page.evaluate(() => {
      const a = document.querySelector('a[href="/Genre/Movie"]');
      return !!a;
    });

    return isMovie ? KissAsianScraper.MEDIA_TYPES.MOVIE : KissAsianScraper.MEDIA_TYPES.SHOW;
  }
}

KissAsianScraper.BASE_URL = 'https://kissasian.li';

KissAsianScraper.DEFAULTS = {
  SITE_SELECTOR: '#head a.logo'
};

KissAsianScraper.MEDIA_TYPES = {
  MOVIE: 'movie',
  SHOW: 'show'
};

module.exports = KissAsianScraper;
